/*
 * ImageKeyframing.java
 */

//TODO: disable growth zones near del. zones

import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;
import java.awt.*;
import java.util.*;

public class ImageKeyframing extends TriangleRegions {

	public ImageKeyframing(int x, int y) {
		super(x, y);
	}
	
	public void fillFromImage(String fname) {
		state = loadImage(fname);
	}
	
	public int[][] loadImage(String fname) {
		int[][] d = new int[maxX][maxY];
		try {
			BufferedImage bi = ImageIO.read(new File(fname));
			Raster r = bi.getData();
			int numBands = r.getNumBands();
			if ((r.getHeight() != maxY) || (r.getWidth() != maxX))
			{
				System.out.println("image file " + fname + " is wrong size.");
				return null;
			}
			for (int i = 0; i < maxX; i++)
				for (int j = 0; j < maxY; j++)
				{
					if (r.getSample(i,j,0) == 0) {
						d[i][j] = FILLED;
					}
					else
					{
						d[i][j] = VOID;
					}
				}
			repaint();
		} catch (IOException e)
		{
			System.out.println("couldn't read image file:" + fname);
			return null;
		}
		return d;
	}
	
	public void regionsFromImage(String fname) {
		int [][] frameB = loadImage(fname);
		
		//compute delta
		for (int i = 0;i < maxX; i++)
			for (int j = 0; j < maxY; j++)
			{
				if (isVoid(new Point(i,j)))
				{
					if (frameB[i][j] == 0) //both void
					{ frameB[i][j] = VOID; }
					else //original is void, B is not
					{ frameB[i][j] = TO_FILL;}
				}
				else
				{
					if (frameB[i][j] == 0) //original is filled, B is not
					{ frameB[i][j] = TO_DELETE; }
					else //both filled
					{ frameB[i][j] = FILLED;}
				}
			}
		
		//create triangles
		createTriRegions();
		//determine region fill state
		determineFillStates(frameB);
		
		//set up inside/outside field
		targetImageArray = loadImage(fname);
		insideTestEnabled = true;
		
		//determine region normal
		setRegionOrigins(frameB);
		
		//set up distanceField
		setDistanceField(frameB);
		
		//state = frameB; //for debugging only!!!
	}
	
	//TODO: implement me!!!
	public void setRegionOrigins(int[][] frameB) {
		//save old state,move frameB in
		int[][] oldState = state;
		state = frameB;
		
		//restore old state
		state = oldState;
	}
	
	public Point perimeterCentroid(TriRegion tr,int[][]frameB) {
		Point min = tr.r.minPt();
		Point max = tr.r.maxPt();
		int xSum,ySum,xCount,yCount;
		xSum = ySum = xCount = yCount = 0;
		for (int i = min.x; i < max.x; i += radius)
			for (int j = min.y; j < max.y; j += radius) {
				
				if (!tr.ptInside(i,j)) continue; //skip if not in triangle
				Point p = screenToArray(i,j);
				if (isPerimeter(p)) {
					xSum += p.x; xCount++;
					ySum += p.y; yCount++;
				}
			}
		return new Point(xSum/xCount,ySum/yCount);
	}
	
	public void fillRegion(TriRegion tr, int fill) {
		Point min = tr.r.minPt();
		Point max = tr.r.maxPt();
		for (int i = min.x; i < max.x; i += radius)
			for (int j = min.y; j < max.y; j += radius) {
				if (!tr.ptInside(i,j)) continue; //skip if not in triangle
				Point p = screenToArray(i,j);
				set(p,fill);
			}
	}
	
	public void determineFillStates(int[][] frameB) {
		Vector regionsToKill = new Vector();
		
		for (int k = 0; k < regions.size(); k++) {
			TriRegion tr = (TriRegion) regions.get(k);
			Point min = tr.r.minPt();
			Point max = tr.r.maxPt();
			int staticCount = 0;
			int growCount = 0;
			int delCount = 0;
			for (int i = min.x; i < max.x; i += radius)
				for (int j = min.y; j < max.y; j += radius) {
					if (!tr.ptInside(i,j)) continue; //skip if not in triangle
					Point p = screenToArray(i,j);
					switch(frameB[p.x][p.y]){
						case TO_FILL:
							growCount++;
							break;
						case TO_DELETE:
							delCount++;
							break;
						default:
							staticCount++;
					}
				}
			if (growCount+delCount <= inactiveThreshold) {
				regionsToKill.add(tr);
			}
			else if (growCount > delCount) {
				tr.fill = true;
			}
			else {
				tr.fill = false;
			}
		}
		//if any regions need to be removed, kill them
		for (int i = 0; i < regionsToKill.size(); i++) {
			regions.remove(regionsToKill.get(i));
		}
	}
	
	public void createTriRegions() {
		for (int i = 0; i < maxX - triangleSize;i+= triangleSize)
			for (int j = 0; j < maxY - 2*triangleSize; j += 2*triangleSize)
			{
				int xMid = i + (int)Math.round(Math.ceil(triangleSize / 2.0)) -1;
				//System.out.println(i + " " + xMid + " " + j);
				TriRegion tr = new TriRegion();
				tr.r = new Triangle();
				tr.r.a = arrayToScreen(i,j);
				tr.r.b = arrayToScreen(i+triangleSize,j);
				tr.r.c = arrayToScreen(xMid,j+triangleSize);
				tr.triX = 2*i/triangleSize;
				tr.triY = j/triangleSize;
				regions.add(tr);
				
				tr = new TriRegion();
				tr.r = new Triangle();
				tr.r.c = arrayToScreen(i,j+2*triangleSize);
				tr.r.b = arrayToScreen(i+triangleSize,j+2*triangleSize);
				tr.r.a = arrayToScreen(xMid,j+triangleSize);
				tr.triX = 2*i/triangleSize;
				tr.triY = j/triangleSize + 1;
				regions.add(tr);
				
				tr = new TriRegion();
				tr.r = new Triangle();
				tr.r.c = arrayToScreen(i+triangleSize,j+2*triangleSize);
				tr.r.b = arrayToScreen(xMid+triangleSize,j+triangleSize);
				tr.r.a = arrayToScreen(xMid,j+triangleSize);
				tr.triX = 2*i/triangleSize + 1;
				tr.triY = j/triangleSize + 1;
				regions.add(tr);
				
				tr = new TriRegion();
				tr.r = new Triangle();
				tr.r.a = arrayToScreen(i+triangleSize,j);
				tr.r.b = arrayToScreen(xMid+triangleSize,j+triangleSize);
				tr.r.c = arrayToScreen(xMid,j+triangleSize);
				tr.triX = 2*i/triangleSize + 1;
				tr.triY = j/triangleSize;
				regions.add(tr);
			}
	}
	
	public void setDistanceField (int frameB[][]) {
		//set initial distances
		for (int i = 0; i < maxX; i++)
			for (int j = 0; j  <maxY; j++) {
				switch(frameB[i][j]) {
					case FILLED:
						distanceField[i][j] = 0;
						break;
					default:
						distanceField[i][j] = -1; 
				}
			}
		
		//set distance field w/ worklist algorithm
		Vector wl = new Vector();
		for (int i = 0; i < maxX; i++)
			for (int j = 0; j  <maxY; j++) {
				if (distanceField[i][j] == 0) {
					for (int k = 1; k < 7; k++) {
						Point p = move(new Point(i,j),k);
						if (!outOfBounds(p) && !wl.contains(p)) wl.add(p);
					}
				}
			}
		
		while (wl.size() > 0) {
			Point p = (Point)wl.remove(0);
			//process point, if its df value goes down, add neighbors to wl
			boolean changed = false;
			for (int k = 1; k < 7; k++) {
				Point p2 = move(p,k);
				if ((!outOfBounds(p2)) &&
				   (distanceField[p2.x][p2.y] != -1) &&
				   ((distanceField[p2.x][p2.y] + 1 < distanceField[p.x][p.y]) ||
				   		(distanceField[p.x][p.y] == -1))) {
					distanceField[p.x][p.y] = distanceField[p2.x][p2.y] + 1;
					changed = true;
				}
			}
			
			if (changed) {
				for (int k = 1; k < 7; k++) {
					Point p2 = move(p,k);
					if ((!outOfBounds(p2)) && (!wl.contains(p2)))
						wl.add(p2);
				}
			}
		}
		
		calcActiveDistanceMax();
	}
	
	int newDistance(int x, int y) {
		int ret = -1;
		Point p = new Point (x,y);
		for (int i = 1; i < 7; i++) {
			Point p2 = move(p,i);
			if (outOfBounds(p2)) continue;
			if (distanceField[p2.x][p2.y] > -1) {
				if (ret == -1) ret = distanceField[p2.x][p2.y]+1;
				else if (ret > distanceField[p2.x][p2.y]+1)
					ret = distanceField[p2.x][p2.y]+1;
			}
		}
		return ret;
	}
	
	void calcActiveDistanceMax() {
		distanceMax = 0;
		for (int i = 0; i < maxX; i++)
			for (int j = 0; j < maxY; j++) {
				if (!isVoid(new Point(i,j)) && (distanceField[i][j] > distanceMax)) {
					distanceMax = distanceField[i][j];
				}
			}
		origDistanceMax = distanceMax;
		System.out.println("distance max " + distanceMax);
	}
	
	void setDistanceMax(int step) {
		distanceMax = origDistanceMax - (step/decayTime);
		if (distanceMax < 0) distanceMax = 0;
		//System.out.println("distance max " + distanceMax);
	}
	
	public void createHoles() {
		int numHoles = (int)Math.round(holeDensity * maxX * maxY);
		for (int i = 0; i < numHoles; i++)
		{
			Hole h = makeHole(new Point(_r.nextInt(maxX), _r.nextInt(maxY)));
			if (h != null) holes.add(h);
		}
	}
	
	public double completionPct() {
		int goodCnt = 0;
		int badCnt = 0;
		for (int i = 0; i < maxX; i++)
			for (int j = 0; j < maxY; j++) {
				if (targetArray[i][j] == VOID) {
					if (state[i][j] == VOID) goodCnt++;
						else badCnt++;
				}
				else {
					if (state[i][j] != VOID) goodCnt++;
						else badCnt++;
				}
			}
		//System.out.println(goodCnt + " " + badCnt);
		return goodCnt/(goodCnt+badCnt*1.0) * 100.0;
	}
	
	public void step() {
		if (ticks % decayTime == 0) setDistanceMax(ticks);
		super.step();
		if ((targetArray != null) && (ticks % 30 == 0))
			System.out.println("% done: " + completionPct());
	}
	
	public void init() {
		super.init();
		if (imageName != null) {
				fillFromImage(imageName);
				createHoles();
		}
		if (targetImage != null) {
			regionsFromImage(targetImage);
			targetArray = loadImage(targetImage);
			repaint();
		}
	}
	
	public static void main(String[] args) {
		ImageKeyframing ik = new ImageKeyframing(100, 100);
		ik.radius = 2;
		ik.setSize(410, 410);
		ik.saveFrames = true;
		ik.sleepTime = 100;
		ik.imageName = "foo.png";
		ik.targetImage = "foo2.png";
		ik.holeDensity = 0.05f;
		ik.mainLoop();
	}
	
	int origDistanceMax;
	
	String imageName = null;
	String targetImage = null;
	int[][] targetArray = null;
	float holeDensity = 0.01f;
	int triangleSize = 15;
	int inactiveThreshold = 5; //number of active pixels in a region before it's used
	int decayTime = 25;
	
	public static final int TO_FILL = 14;
	public static final int TO_DELETE = 28;
	public static final int EDGE = 19;
}
